/*
 * Copyright 2009-2013 by The Regents of the University of California
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * you may obtain a copy of the License from
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package edu.uci.ics.asterix.runtime.operators.file;

import java.io.DataOutput;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import edu.uci.ics.asterix.builders.IARecordBuilder;
import edu.uci.ics.asterix.builders.RecordBuilder;
import edu.uci.ics.asterix.common.exceptions.AsterixException;
import edu.uci.ics.asterix.dataflow.data.nontagged.serde.ANullSerializerDeserializer;
import edu.uci.ics.asterix.om.base.AMutableString;
import edu.uci.ics.asterix.om.base.ANull;
import edu.uci.ics.asterix.om.types.ARecordType;
import edu.uci.ics.asterix.om.types.ATypeTag;
import edu.uci.ics.asterix.om.types.AUnionType;
import edu.uci.ics.asterix.om.util.NonTaggedFormatUtil;
import edu.uci.ics.hyracks.api.exceptions.HyracksDataException;
import edu.uci.ics.hyracks.data.std.util.ArrayBackedValueStorage;
import edu.uci.ics.hyracks.dataflow.common.data.marshalling.Integer64SerializerDeserializer;
import edu.uci.ics.hyracks.dataflow.common.data.parsers.IValueParser;
import edu.uci.ics.hyracks.dataflow.common.data.parsers.IValueParserFactory;
import edu.uci.ics.hyracks.dataflow.std.file.FieldCursorForDelimitedDataParser;

public class DelimitedDataParser extends AbstractDataParser implements IDataParser {

    protected final IValueParserFactory[] valueParserFactories;
    protected final char fieldDelimiter;
    protected final char quote;
    protected final boolean hasHeader;
    protected final ARecordType recordType;

    private IARecordBuilder recBuilder;
    private ArrayBackedValueStorage fieldValueBuffer;
    private DataOutput fieldValueBufferOutput;
    private IValueParser[] valueParsers;
    private FieldCursorForDelimitedDataParser cursor;
    private byte[] fieldTypeTags;
    private int[] fldIds;
    private ArrayBackedValueStorage[] nameBuffers;
    private byte AUUIDTag = ATypeTag.UUID.serialize();

    // Variables used to set a UUID for the auto-generated PK field
    private final boolean isPKAutoGenerated;
    private final int primaryKeyPosition;
    private final ARecordType origRecordTypeForAutoGeneratedPK;

    private boolean areAllNullFields;
    private int fieldCount;

    public DelimitedDataParser(ARecordType recordType, IValueParserFactory[] valueParserFactories, char fieldDelimter,
            char quote, boolean hasHeader) {
        this(recordType, valueParserFactories, fieldDelimter, quote, hasHeader, false, -1, null);
    }

    public DelimitedDataParser(ARecordType recordType, IValueParserFactory[] valueParserFactories, char fieldDelimter,
            char quote, boolean hasHeader, boolean isPKAutoGenerated, int primaryKeyPosition,
            ARecordType origRecordTypeForAutoGeneratedPK) {
        this.recordType = recordType;
        this.valueParserFactories = valueParserFactories;
        this.fieldDelimiter = fieldDelimter;
        this.quote = quote;
        this.hasHeader = hasHeader;
        this.isPKAutoGenerated = isPKAutoGenerated;
        this.primaryKeyPosition = primaryKeyPosition;
        this.origRecordTypeForAutoGeneratedPK = origRecordTypeForAutoGeneratedPK;
    }

    @Override
    public void initialize(InputStream in, ARecordType recordType, boolean datasetRec) throws AsterixException,
            IOException {

        ARecordType recordTypeToApply = null;
        if (isPKAutoGenerated)
            recordTypeToApply = origRecordTypeForAutoGeneratedPK;
        else
            recordTypeToApply = recordType;

        valueParsers = new IValueParser[valueParserFactories.length];
        for (int i = 0; i < valueParserFactories.length; ++i) {
            valueParsers[i] = valueParserFactories[i].createValueParser();
        }

        fieldValueBuffer = new ArrayBackedValueStorage();
        fieldValueBufferOutput = fieldValueBuffer.getDataOutput();

        // If PK is auto-generated, then we need to use the recordType that
        // includes PK, since recordType variable does not include PK field.
        recBuilder = new RecordBuilder();
        recBuilder.reset(recordTypeToApply);
        recBuilder.init();

        int n = recordType.getFieldNames().length;
        fieldTypeTags = new byte[n];
        for (int i = 0; i < n; i++) {
            ATypeTag tag = recordType.getFieldTypes()[i].getTypeTag();
            fieldTypeTags[i] = tag.serialize();
        }

        fldIds = new int[n];
        nameBuffers = new ArrayBackedValueStorage[n];
        AMutableString str = new AMutableString(null);
        for (int i = 0; i < n; i++) {
            String name = recordType.getFieldNames()[i];
            fldIds[i] = recBuilder.getFieldId(name);
            if (fldIds[i] < 0) {
                if (!recordType.isOpen()) {
                    throw new HyracksDataException("Illegal field " + name + " in closed type " + recordType);
                } else {
                    nameBuffers[i] = new ArrayBackedValueStorage();
                    fieldNameToBytes(name, str, nameBuffers[i]);
                }
            }
        }

        cursor = new FieldCursorForDelimitedDataParser(new InputStreamReader(in), fieldDelimiter, quote);
    }

    @Override
    public boolean parse(DataOutput out) throws AsterixException, IOException {
        if (hasHeader && cursor.lineCount == 1) {
            // Consume all fields of first record
            cursor.nextRecord();
            while (cursor.nextField(fieldCount));
        }
        while (cursor.nextRecord()) {
            // If PK is auto-generated, then we need to use the recordType that
            // includes PK, since recordType variable does not include PK field.
            if (isPKAutoGenerated)
                recBuilder.reset(origRecordTypeForAutoGeneratedPK);
            else
                recBuilder.reset(recordType);

            recBuilder.init();
            areAllNullFields = true;

            fieldCount = 0;

            for (int i = 0; i < valueParsers.length; ++i) {
                if (!cursor.nextField(fieldCount)) {
                    break;
                }
                fieldValueBuffer.reset();

                if (cursor.fStart == cursor.fEnd && recordType.getFieldTypes()[i].getTypeTag() != ATypeTag.STRING
                        && recordType.getFieldTypes()[i].getTypeTag() != ATypeTag.NULL) {
                    // if the field is empty and the type is optional, insert
                    // NULL. Note that string type can also process empty field as an
                    // empty string
                    if (recordType.getFieldTypes()[i].getTypeTag() != ATypeTag.UNION
                            || !NonTaggedFormatUtil.isOptionalField((AUnionType) recordType.getFieldTypes()[i])) {
                        throw new AsterixException("At line: " + cursor.lineCount + " - Field " + i
                                + " is not an optional type so it cannot accept null value. ");
                    }
                    fieldValueBufferOutput.writeByte(ATypeTag.NULL.serialize());
                    ANullSerializerDeserializer.INSTANCE.serialize(ANull.NULL, out);
                } else {
                    fieldValueBufferOutput.writeByte(fieldTypeTags[i]);
                    // Eliminate doule quotes in the field that we are going to parse
                    if (cursor.isDoubleQuoteIncludedInThisField) {
                        cursor.eliminateDoubleQuote(cursor.buffer, cursor.fStart, cursor.fEnd - cursor.fStart);
                        cursor.fEnd -= cursor.doubleQuoteCount;
                        cursor.isDoubleQuoteIncludedInThisField = false;
                    }
                    valueParsers[i].parse(cursor.buffer, cursor.fStart, cursor.fEnd - cursor.fStart,
                            fieldValueBufferOutput);
                    areAllNullFields = false;
                }
                if (fldIds[i] < 0) {
                    recBuilder.addField(nameBuffers[i], fieldValueBuffer);
                } else {
                    recBuilder.addField(fldIds[i], fieldValueBuffer);
                }

                fieldCount++;

            }

            // Should not have any more fields now.
            // We parsed all fields except an auto-generated PK at this point.
            // We now create a new UUID and assign it as a PK.
            if (isPKAutoGenerated && fieldCount == origRecordTypeForAutoGeneratedPK.getFieldTypes().length - 1) {
                fieldValueBuffer.reset();
                aUUID.nextUUID();
                fieldValueBufferOutput.writeByte(AUUIDTag);
                Integer64SerializerDeserializer.INSTANCE.serialize(aUUID.getMostSignificantBits(),
                        fieldValueBufferOutput);
                Integer64SerializerDeserializer.INSTANCE.serialize(aUUID.getLeastSignificantBits(),
                        fieldValueBufferOutput);
                recBuilder.addField(primaryKeyPosition, fieldValueBuffer);
                areAllNullFields = false;
            } else if (isPKAutoGenerated && fieldCount >= origRecordTypeForAutoGeneratedPK.getFieldTypes().length) {
                // If we have all fields in the file including auto-generated PK,
                // throw an exception
                throw new AsterixException("At line: " + cursor.lineCount
                        + " - check number of fields. Auto-generated PK field should not exist in the input data.");
            }

            if (!areAllNullFields) {
                recBuilder.write(out, true);
                return true;
            }
        }
        return false;
    }

    protected void fieldNameToBytes(String fieldName, AMutableString str, ArrayBackedValueStorage buffer)
            throws HyracksDataException {
        buffer.reset();
        DataOutput out = buffer.getDataOutput();
        str.setValue(fieldName);
        try {
            stringSerde.serialize(str, out);
        } catch (IOException e) {
            throw new HyracksDataException(e);
        }
    }

}
